@@ -245,6 +245,78 @@ void DBusMenuExporterPrivate::insertIconProperty(QVariantMap *map, QAction *acti
     // "icon-data";
 }
 
+static void collapseSeparator(QAction* action)
+{
+    action->setVisible(false);
+}
+
+// Unless the separatorsCollapsible property is set to false, Qt will get rid
+// of separators at the beginning and at the end of menus as well as collapse
+// multiple separators in the middle. For example, a menu like this:
+//
+// ---
+// Open
+// ---
+// ---
+// Quit
+// ---
+//
+// is displayed like this:
+//
+// Open
+// ---
+// Quit
+//
+// We fake this by setting separators invisible before exporting them.
+//
+// cf. https://bugs.launchpad.net/libdbusmenu-qt/+bug/793339
+void DBusMenuExporterPrivate::collapseSeparators(QMenu* menu)
+{
+    QList<QAction*> actions = menu->actions();
+    if (actions.isEmpty()) {
+        return;
+    }
+
+    QList<QAction*>::Iterator it, begin = actions.begin(), end = actions.end();
+
+    // Get rid of separators at end
+    it = end - 1;
+    for (; it != begin; --it) {
+        if ((*it)->isSeparator()) {
+            collapseSeparator(*it);
+        } else {
+            break;
+        }
+    }
+    // end now points after the last visible entry
+    end = it + 1;
+    it = begin;
+
+    // Get rid of separators at beginnning
+    for (; it != end; ++it) {
+        if ((*it)->isSeparator()) {
+            collapseSeparator(*it);
+        } else {
+            break;
+        }
+    }
+
+    // Collapse separators in between
+    bool previousWasSeparator = false;
+    for (; it != end; ++it) {
+        QAction* action = *it;
+        if (action->isSeparator()) {
+            if (previousWasSeparator) {
+                collapseSeparator(action);
+            } else {
+                previousWasSeparator = true;
+            }
+        } else {
+            previousWasSeparator = false;
+        }
+    }
+}
+
 //-------------------------------------------------
 //
 // DBusMenuExporter
@@ -255,6 +327,7 @@ DBusMenuExporter::DBusMenuExporter(const QString &objectPath, QMenu *menu, const
 , d(new DBusMenuExporterPrivate)
 {
     d->q = this;
+    d->m_objectPath = objectPath;
     d->m_rootMenu = menu;
     d->m_nextId = 1;
     d->m_revision = 1;
@@ -348,6 +421,10 @@ void DBusMenuExporter::doUpdateActions()
 void DBusMenuExporter::doEmitLayoutUpdated()
 {
     Q_FOREACH(int id, d->m_layoutUpdatedIds) {
+        QMenu* menu = d->menuForId(id);
+        if (menu && menu->separatorsCollapsible()) {
+            d->collapseSeparators(menu);
+        }
         d->m_dbusObject->LayoutUpdated(d->m_revision, id);
     }
     d->m_layoutUpdatedIds.clear();
@@ -358,7 +435,11 @@ QString DBusMenuExporter::iconNameForAction(QAction *action)
     DMRETURN_VALUE_IF_FAIL(action, QString());
 #ifdef HAVE_QICON_NAME
     QIcon icon = action->icon();
-    return icon.isNull() ? QString() : icon.name();
+    if (action->isIconVisibleInMenu() && !icon.isNull()) {
+        return icon.name();
+    } else {
+        return QString();
+    }
 #else
     return QString();
 #endif
@@ -377,4 +458,14 @@ void DBusMenuExporter::slotActionDestroyed(QObject* object)
     d->removeActionInternal(object);
 }
 
+void DBusMenuExporter::setStatus(const QString& status)
+{
+    d->m_dbusObject->setStatus(status);
+}
+
+QString DBusMenuExporter::status() const
+{
+    return d->m_dbusObject->status();
+}
+
 #include "dbusmenuexporter.moc"
