@@ -40,6 +40,8 @@ class DBusMenuExporterPrivate
 public:
     DBusMenuExporter *q;
 
+    QString m_objectPath;
+
     DBusMenuExporterDBus *m_dbusObject;
 
     QMenu *m_rootMenu;
@@ -80,6 +82,8 @@ public:
     void emitLayoutUpdated(int id);
 
     void insertIconProperty(QVariantMap* map, QAction *action) const;
+
+    void collapseSeparators(QMenu*);
 };
 
 
