@@ -31,7 +31,7 @@
 
 // DBusMenuQt
 #include <dbusmenuexporter.h>
-#include <dbusmenuitem_p.h>
+#include <dbusmenutypes_p.h>
 #include <dbusmenushortcut_p.h>
 #include <debug_p.h>
 
@@ -45,9 +45,23 @@ static const char *TEST_OBJECT_PATH = "/TestMenuBar";
 
 Q_DECLARE_METATYPE(QList<int>)
 
+static DBusMenuLayoutItemList getChildren(QDBusAbstractInterface* iface, int parentId, const QStringList &propertyNames)
+{
+    QDBusPendingReply<uint, DBusMenuLayoutItem> reply = iface->call("GetLayout", parentId, /*recursionDepth=*/ 1, propertyNames);
+    reply.waitForFinished();
+    if (!reply.isValid()) {
+        qFatal("%s", qPrintable(reply.error().message()));
+        return DBusMenuLayoutItemList();
+    }
+
+    DBusMenuLayoutItem rootItem = reply.argumentAt<1>();
+    return rootItem.children;
+}
+
 void DBusMenuExporterTest::init()
 {
     QVERIFY(QDBusConnection::sessionBus().registerService(TEST_SERVICE));
+    QCoreApplication::setAttribute(Qt::AA_DontShowIconsInMenus, false);
 }
 
 void DBusMenuExporterTest::cleanup()
@@ -91,15 +105,10 @@ void DBusMenuExporterTest::testGetSomeProperties()
 
     // Get exported menu info
     QStringList propertyNames = QStringList() << "type" << "enabled" << "label" << "icon-name";
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, propertyNames);
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    // Check the info we received, in particular, check that any property set to
-    // its default value is *not* exported
-    DBusMenuItemList list = reply.value();
-    QCOMPARE(list.count(), 1);
-    DBusMenuItem item = list.first();
+    DBusMenuLayoutItemList list = getChildren(&iface, /*parentId=*/0, propertyNames);
+    DBusMenuLayoutItem item = list.first();
     QVERIFY(item.id != 0);
+    QVERIFY(item.children.isEmpty());
     QVERIFY(!item.properties.contains("type"));
     QCOMPARE(item.properties.value("label").toString(), label);
     if (enabled) {
@@ -153,14 +162,11 @@ void DBusMenuExporterTest::testGetAllProperties()
     QVERIFY2(iface.isValid(), qPrintable(iface.lastError().message()));
 
     // Get children
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 3);
 
     // Check we get the right properties
-    DBusMenuItem item = list.takeFirst();
+    DBusMenuLayoutItem item = list.takeFirst();
     QCOMPARE(QSet<QString>::fromList(item.properties.keys()), a1Properties);
 
     item = list.takeFirst();
@@ -179,13 +185,10 @@ void DBusMenuExporterTest::testGetNonExistentProperty()
     DBusMenuExporter exporter(TEST_OBJECT_PATH, &inputMenu);
 
     QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList() << NON_EXISTENT_KEY);
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList() << NON_EXISTENT_KEY);
     QCOMPARE(list.count(), 1);
 
-    DBusMenuItem item = list.takeFirst();
+    DBusMenuLayoutItem item = list.takeFirst();
     QVERIFY(!item.properties.contains(NON_EXISTENT_KEY));
 }
 
@@ -197,10 +200,7 @@ void DBusMenuExporterTest::testClickedEvent()
     DBusMenuExporter exporter(TEST_OBJECT_PATH, &inputMenu);
 
     QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 1);
     int id = list.first().id;
 
@@ -221,19 +221,14 @@ void DBusMenuExporterTest::testSubMenu()
     DBusMenuExporter exporter(TEST_OBJECT_PATH, &inputMenu);
 
     QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 1);
     int id = list.first().id;
 
-    reply = iface.call("GetChildren", id, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    list = reply.value();
+    list = getChildren(&iface, id, QStringList());
     QCOMPARE(list.count(), 2);
 
-    DBusMenuItem item = list.takeFirst();
+    DBusMenuLayoutItem item = list.takeFirst();
     QVERIFY(item.id != 0);
     QCOMPARE(item.properties.value("label").toString(), a1->text());
 
@@ -260,9 +255,7 @@ void DBusMenuExporterTest::testDynamicSubMenu()
     filler.addAction(new QAction("a2", subMenu));
 
     // Get id of submenu
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 1);
     int id = list.first().id;
 
@@ -276,21 +269,19 @@ void DBusMenuExporterTest::testDynamicSubMenu()
 
     // Pretend we show the menu
     QDBusReply<bool> aboutToShowReply = iface.call("AboutToShow", id);
-    QVERIFY2(aboutToShowReply.isValid(), qPrintable(reply.error().message()));
+    QVERIFY2(aboutToShowReply.isValid(), qPrintable(aboutToShowReply.error().message()));
     QVERIFY(aboutToShowReply.value());
     QTest::qWait(500);
     QCOMPARE(layoutUpdatedSpy.count(), 1);
     QCOMPARE(layoutUpdatedSpy.takeFirst().at(1).toInt(), id);
 
     // Get submenu items
-    reply = iface.call("GetChildren", id, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    list = reply.value();
+    list = getChildren(&iface, id, QStringList());
     QVERIFY(subMenu->actions().count() > 0);
     QCOMPARE(list.count(), subMenu->actions().count());
 
     for (int pos=0; pos< list.count(); ++pos) {
-        DBusMenuItem item = list.at(pos);
+        DBusMenuLayoutItem item = list.at(pos);
         QVERIFY(item.id != 0);
         QAction *action = subMenu->actions().at(pos);
         QVERIFY(action);
@@ -300,8 +291,8 @@ void DBusMenuExporterTest::testDynamicSubMenu()
 
 void DBusMenuExporterTest::testRadioItems()
 {
-    DBusMenuItem item;
-    DBusMenuItemList list;
+    DBusMenuLayoutItem item;
+    DBusMenuLayoutItemList list;
     QMenu inputMenu;
     QVERIFY(QDBusConnection::sessionBus().registerService(TEST_SERVICE));
     DBusMenuExporter exporter(TEST_OBJECT_PATH, &inputMenu);
@@ -321,9 +312,7 @@ void DBusMenuExporterTest::testRadioItems()
 
     // Get item ids
     QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    list = reply.value();
+    list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 2);
 
     // Check items are radios and correctly toggled
@@ -338,16 +327,15 @@ void DBusMenuExporterTest::testRadioItems()
 
     // Click a2
     ManualSignalSpy spy;
-    QDBusConnection::sessionBus().connect(TEST_SERVICE, TEST_OBJECT_PATH, "com.canonical.dbusmenu", "ItemUpdated", &spy, SLOT(receiveCall(int)));
+    QDBusConnection::sessionBus().connect(TEST_SERVICE, TEST_OBJECT_PATH, "com.canonical.dbusmenu", "ItemsPropertiesUpdated", "a(ia{sv})a(ias)",
+        &spy, SLOT(receiveCall(DBusMenuItemList, DBusMenuItemKeysList)));
     QVariant empty = QVariant::fromValue(QDBusVariant(QString()));
     uint timestamp = QDateTime::currentDateTime().toTime_t();
     iface.call("Event", a2Id, "clicked", empty, timestamp);
     QTest::qWait(500);
 
     // Check a1 is not checked, but a2 is
-    reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    list = reply.value();
+    list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 2);
 
     item = list.takeFirst();
@@ -357,10 +345,14 @@ void DBusMenuExporterTest::testRadioItems()
     QCOMPARE(item.properties.value("toggle-state").toInt(), 1);
 
     // Did we get notified?
-    QCOMPARE(spy.count(), 2);
+    QCOMPARE(spy.count(), 1);
     QSet<int> updatedIds;
-    updatedIds << spy.takeFirst().at(0).toInt();
-    updatedIds << spy.takeFirst().at(0).toInt();
+    {
+        QVariantList lst = spy.takeFirst().at(0).toList();
+        Q_FOREACH(QVariant variant, lst) {
+            updatedIds << variant.toInt();
+        }
+    }
 
     QSet<int> expectedIds;
     expectedIds << a1Id << a2Id;
@@ -370,8 +362,8 @@ void DBusMenuExporterTest::testRadioItems()
 
 void DBusMenuExporterTest::testNonExclusiveActionGroup()
 {
-    DBusMenuItem item;
-    DBusMenuItemList list;
+    DBusMenuLayoutItem item;
+    DBusMenuLayoutItemList list;
     QMenu inputMenu;
     QVERIFY(QDBusConnection::sessionBus().registerService(TEST_SERVICE));
     DBusMenuExporter exporter(TEST_OBJECT_PATH, &inputMenu);
@@ -390,9 +382,7 @@ void DBusMenuExporterTest::testNonExclusiveActionGroup()
 
     // Get item ids
     QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    list = reply.value();
+    list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 2);
 
     // Check items are checkmark, not radio
@@ -414,9 +404,7 @@ void DBusMenuExporterTest::testClickDeletedAction()
 
     // Get id
     QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), 1);
     int id = list.takeFirst().id;
 
@@ -486,16 +474,12 @@ void DBusMenuExporterTest::testMenuShortcut()
     QVERIFY2(iface.isValid(), qPrintable(iface.lastError().message()));
 
     // Get exported menu info
-    QStringList propertyNames = QStringList() << "shortcut";
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, propertyNames);
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    // Check the info we received
-    DBusMenuItemList list = reply.value();
+    QStringList propertyNames = QStringList() << "label" << "shortcut";
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, propertyNames);
     QCOMPARE(list.count(), actionList.count());
 
     Q_FOREACH(const QAction* action, actionList) {
-        DBusMenuItem item = list.takeFirst();
+        DBusMenuLayoutItem item = list.takeFirst();
         if (action->shortcut().isEmpty()) {
             QVERIFY(!item.properties.contains("shortcut"));
         } else {
@@ -523,10 +507,7 @@ void DBusMenuExporterTest::testGetGroupProperties()
     QVERIFY2(iface.isValid(), qPrintable(iface.lastError().message()));
 
     // Get item ids
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), inputMenu.actions().count());
 
     int id1 = list.at(0).id;
@@ -534,15 +515,15 @@ void DBusMenuExporterTest::testGetGroupProperties()
 
     // Get group properties
     QList<int> ids = QList<int>() << id1 << id2;
-    reply = iface.call("GetGroupProperties", QVariant::fromValue(ids), QStringList());
+    QDBusReply<DBusMenuItemList> reply = iface.call("GetGroupProperties", QVariant::fromValue(ids), QStringList());
     QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-    list = reply.value();
+    DBusMenuItemList groupPropertiesList = reply.value();
 
     // Check the info we received
-    QCOMPARE(list.count(), inputMenu.actions().count());
+    QCOMPARE(groupPropertiesList.count(), inputMenu.actions().count());
 
     Q_FOREACH(const QAction* action, inputMenu.actions()) {
-        DBusMenuItem item = list.takeFirst();
+        DBusMenuItem item = groupPropertiesList.takeFirst();
         QCOMPARE(item.properties.value("label").toString(), action->text());
     }
 }
@@ -565,10 +546,7 @@ void DBusMenuExporterTest::testActivateAction()
     QDBusConnection::sessionBus().connect(TEST_SERVICE, TEST_OBJECT_PATH, "com.canonical.dbusmenu", "ItemActivationRequested", "iu", &spy, SLOT(receiveCall(int, uint)));
 
     // Get item ids
-    QDBusReply<DBusMenuItemList> reply = iface.call("GetChildren", 0, QStringList());
-    QVERIFY2(reply.isValid(), qPrintable(reply.error().message()));
-
-    DBusMenuItemList list = reply.value();
+    DBusMenuLayoutItemList list = getChildren(&iface, 0, QStringList());
     QCOMPARE(list.count(), inputMenu.actions().count());
 
     int id1 = list.at(0).id;
@@ -622,4 +600,167 @@ void DBusMenuExporterTest::testTrackActionsOnlyOnce()
     QCOMPARE(trackCount(subMenu), 1);
 }
 
+// If desktop does not want icon in menus, check we do not export them
+void DBusMenuExporterTest::testHonorDontShowIconsInMenusAttribute()
+{
+    QCoreApplication::setAttribute(Qt::AA_DontShowIconsInMenus, true);
+    QMenu inputMenu;
+    DBusMenuExporter exporter(TEST_OBJECT_PATH, &inputMenu);
+
+    QAction *action = new QAction("Undo", &inputMenu);
+    QIcon icon = QIcon::fromTheme("edit-undo");
+    QVERIFY(!icon.isNull());
+    action->setIcon(icon);
+    inputMenu.addAction(action);
+
+    // Check out exporter is on DBus
+    QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
+    QVERIFY2(iface.isValid(), qPrintable(iface.lastError().message()));
+
+    // Get exported menu info
+    QStringList propertyNames = QStringList() << "icon-name";
+    DBusMenuLayoutItemList list = getChildren(&iface, /*parentId=*/0, propertyNames);
+    DBusMenuLayoutItem item = list.first();
+    QVERIFY(item.id != 0);
+    QVERIFY(!item.properties.contains("icon-name"));
+}
+
+static bool hasInternalDBusMenuObject(QMenu* menu)
+{
+    Q_FOREACH(QObject* obj, menu->children()) {
+        if (obj->inherits("DBusMenu")) {
+            return true;
+        }
+    }
+    return false;
+}
+
+// DBusMenuExporter adds an instance of an internal class named "DBusMenu" to
+// any QMenu it tracks. Check they go away when the exporter is deleted.
+void DBusMenuExporterTest::testDBusMenuObjectIsDeletedWhenExporterIsDeleted()
+{
+    QMenu inputMenu;
+    QVERIFY(QDBusConnection::sessionBus().registerService(TEST_SERVICE));
+    DBusMenuExporter *exporter = new DBusMenuExporter(TEST_OBJECT_PATH, &inputMenu);
+
+    QAction *a1 = inputMenu.addAction("a1");
+    QVERIFY2(hasInternalDBusMenuObject(&inputMenu), "Test setup failed");
+    delete exporter;
+    QVERIFY(!hasInternalDBusMenuObject(&inputMenu));
+}
+
+void DBusMenuExporterTest::testSeparatorCollapsing_data()
+{
+    QTest::addColumn<QString>("input");
+    QTest::addColumn<QString>("expected");
+
+    QTest::newRow("one-separator")         << "a-b"           << "a-b";
+    QTest::newRow("two-separators")        << "a-b-c"         << "a-b-c";
+    QTest::newRow("middle-separators")     << "a--b"          << "a-b";
+    QTest::newRow("separators-at-begin")   << "--a-b"         << "a-b";
+    QTest::newRow("separators-at-end")     << "a-b--"         << "a-b";
+    QTest::newRow("separators-everywhere") << "--a---bc--d--" << "a-bc-d";
+    QTest::newRow("empty-menu")            << ""              << "";
+    QTest::newRow("separators-only")       << "---"           << "";
+}
+
+void DBusMenuExporterTest::testSeparatorCollapsing()
+{
+    QFETCH(QString, input);
+    QFETCH(QString, expected);
+
+    // Create menu from menu string
+    QMenu inputMenu;
+
+    QVERIFY(QDBusConnection::sessionBus().registerService(TEST_SERVICE));
+    DBusMenuExporter *exporter = new DBusMenuExporter(TEST_OBJECT_PATH, &inputMenu);
+
+    if (input.isEmpty()) {
+        // Pretend there was an action so that doEmitLayoutUpdated() is called
+        // even if the new menu is empty. If we don't do this we don't test
+        // DBusMenuExporterPrivate::collapseSeparators() for empty menus.
+        delete inputMenu.addAction("dummy");
+    }
+
+    Q_FOREACH(QChar ch, input) {
+        if (ch == '-') {
+            inputMenu.addSeparator();
+        } else {
+            inputMenu.addAction(ch);
+        }
+    }
+
+    QTest::qWait(500);
+
+    // Check out exporter is on DBus
+    QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
+    QVERIFY2(iface.isValid(), qPrintable(iface.lastError().message()));
+
+    // Get exported menu info
+    QStringList propertyNames = QStringList();
+    DBusMenuLayoutItemList list = getChildren(&iface, /*parentId=*/0, propertyNames);
+
+    // Recreate a menu string from the item list
+    QString output;
+    Q_FOREACH(const DBusMenuLayoutItem& item, list) {
+        QVariantMap properties = item.properties;
+        if (properties.contains("visible") && !properties.value("visible").toBool()) {
+            continue;
+        }
+        QString type = properties.value("type").toString();
+        if (type == "separator") {
+            output += '-';
+        } else {
+            output += properties.value("label").toString();
+        }
+    }
+
+    // Check it matches
+    QCOMPARE(output, expected);
+}
+
+static void checkPropertiesChangedArgs(const QVariantList& args, const QString& name, const QVariant& value)
+{
+    QCOMPARE(args[0].toString(), QString("com.canonical.dbusmenu"));
+    QVariantMap map;
+    map.insert(name, value);
+    QCOMPARE(args[1].toMap(), map);
+    QCOMPARE(args[2].toStringList(), QStringList());
+}
+
+void DBusMenuExporterTest::testSetStatus()
+{
+    QMenu inputMenu;
+    QVERIFY(QDBusConnection::sessionBus().registerService(TEST_SERVICE));
+    DBusMenuExporter *exporter = new DBusMenuExporter(TEST_OBJECT_PATH, &inputMenu);
+    ManualSignalSpy spy;
+    QDBusConnection::sessionBus().connect(TEST_SERVICE, TEST_OBJECT_PATH, "org.freedesktop.DBus.Properties", "PropertiesChanged", "sa{sv}as", &spy, SLOT(receiveCall(QString, QVariantMap, QStringList)));
+
+    QTest::qWait(500);
+
+    // Check our exporter is on DBus
+    QDBusInterface iface(TEST_SERVICE, TEST_OBJECT_PATH);
+    QVERIFY2(iface.isValid(), qPrintable(iface.lastError().message()));
+
+    QCOMPARE(exporter->status(), QString("normal"));
+
+    // Change status, a DBus signal should be emitted
+    exporter->setStatus("notice");
+    QCOMPARE(exporter->status(), QString("notice"));
+    QTest::qWait(500);
+    QCOMPARE(spy.count(), 1);
+    checkPropertiesChangedArgs(spy.takeFirst(), "Status", "notice");
+
+    // Same status => no signal
+    exporter->setStatus("notice");
+    QTest::qWait(500);
+    QCOMPARE(spy.count(), 0);
+
+    // Change status, a DBus signal should be emitted
+    exporter->setStatus("normal");
+    QTest::qWait(500);
+    QCOMPARE(spy.count(), 1);
+    checkPropertiesChangedArgs(spy.takeFirst(), "Status", "normal");
+}
+
 #include "dbusmenuexportertest.moc"
